@@ -26,8 +26,6 @@ sub init ($self) {
 
     $self->{map} = $map;
 
-    $self->{_unblocked_app} = Coro::unblock_sub sub ( $ctrl, $req ) { $ctrl->run($req); return };
-
     return;
 }
 
@@ -177,7 +175,9 @@ sub run ( $self, $req ) {
     $req->{path}      = $path;
     $req->{path_tail} = P->path($path_tail);
 
-    $self->{_unblocked_app}->( $self->{_path_class_cache}->{$host}->{$path}, $req );
+    my $ctrl = $self->{_path_class_cache}->{$host}->{$path};
+
+    Coro::async_pool { $ctrl->run($req) };
 
     return;
 }
